/**
 * @file
 *
 * @copyright Axiros GmbH 2019-2022, All rights reserved
 *
 * @brief Functionality for frontend interaction definitions
 */

#ifndef CORE_FRONTEND_H_
#define CORE_FRONTEND_H_

#include "config.h"
#include "compiler.h"

#include "core/notifications.h"

/* Forward declarations */
struct dm_object;
struct dm_parameter;
struct event_base;
struct frontend;

/**
 * Definition of a out-of-session-event callback function to be called after an
 * ongoing session has finished */
typedef void (*oose_callback) (struct event_base *base, void *data);


/**
 * @brief Initialises a new empty frontend
 *
 * @param[in] root	the root object to use for the datamodel
 *
 * @return a freshly allocated frontend
 */
struct frontend *frontend_init (struct dm_object *root) MALLOC ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the system config path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure
 */
void frontend_set_sysconfig_path (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the system config path used by this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the path used by this frontend
 */
char *frontend_get_sysconfig_path (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the system state path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure
 */
void frontend_set_sysstate_path (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the system state path used by this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the path used by this frontend
 */
char *frontend_get_sysstate_path (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the log path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure
 */
void frontend_set_log_path (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the log path used by this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the path used by this frontend
 */
char *frontend_get_log_path (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the bootstrap file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_bootstrap_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the bootstrap file configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path to the bootstrap file
 */
char *frontend_get_bootstrap_file (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the database file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_db_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the database file configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path to the internal configuration storage
 */
char *frontend_get_db_file (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the status file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_status_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the status file configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path to the status file
 */
char *frontend_get_status_file (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the metadata file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_metadata_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the metadata file configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path to the metadata file
 */
char *frontend_get_metadata_file (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the log file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_log_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the log file configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path to the general log
 */
char *frontend_get_log_file (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the TLS certificaten bundle file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_ca_certs_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the TLS certificate bundle configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path for the TLS CA certificate bundle
 */
char *frontend_get_ca_certs_file (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Configure the TLS client certificate file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_client_cert_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the TLS client certificate configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path for the TLS client certificate, can be NULL if not global client certificates are used
 */
char *frontend_get_client_cert_file (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Check whether the client certificate file is configured and exists
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return 1 if the file is configured and exists, 0 otherwise
 */
int frontend_client_cert_file_exists (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Configure the TLS client key file path to use by this frontend
 *
 * @param[in] fe	the frontend to configure the path on
 * @param[in] path	the path to configure, must be a heap allocated string!
 */
void frontend_set_client_key_file (struct frontend *fe, char *path) ARGS_NONNULL;


/**
 * @brief Retrieve the path to the TLS client key configured on this frontend
 *
 * @param[in] fe	the frontend to get the configured path from
 *
 * @return the full filename including the path for the TLS client key, can be NULL if not global client certificates are used
 */
char *frontend_get_client_key_file (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Return the runtime of this frontend instance in seconds
 *
 * @param[in] fe	the frontend to use for the runtime calculation
 *
 * @return the number of seconds since initialisation of this frontend instance
 */
unsigned int frontend_get_runtime (struct frontend *fe) ARGS_NONNULL;


#if defined(ENABLE_USP)
/**
 * @brief Use the initialised (and USP capable frontend) to create an impromptu USP OnBoardRequest notification via a MQTT MTP
 *
 * @param[in] fe	the frontend to use for the impromptu OnBoardRequest
 * @param[in] cei	the Endpoint ID of the controller to send the OnBoardRequest to
 * @param[in] url	a valid URL string suitable for a connection to a MQTT broker
 * @param[in] persist	use persistent datamodel objects rather than temporary ones
 */
void frontend_impromptu_mqtt_obr (struct frontend *fe, char *cei, char *url, int persist) ARGS_NONNULL;


/**
 * @brief Use the initialised (and USP capable frontend) to create an impromptu USP OnBoardRequest notification via a Websockets MTP
 *
 * @param[in] fe	the frontend to use for the impromptu OnBoardRequest
 * @param[in] cei	the Endpoint ID of the controller to send the OnBoardRequest to
 * @param[in] url	a valid URL string suitable for a connection Websockets server
 * @param[in] persist	use persistent datamodel objects rather than temporary ones
 */
void frontend_impromptu_websockets_obr (struct frontend *fe, char *cei, char *url, int persist) ARGS_NONNULL;


/**
 * @brief Initialises a USP engine and adds it to the frontend
 */
void frontend_add_usp_engine (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Return the USP engine associated with this frontend
 *
 * @return the associated USP engine or NULL if none available
 */
struct usp_engine *frontend_get_usp_engine (struct frontend *fe) ARGS_NONNULL;
#endif


#if defined(ENABLE_TR069)
/**
 * @brief Initialises a TR-069 engine and adds it to the frontend
 */
void frontend_add_tr069_engine (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Return the TR-069 engine associated with this frontend
 *
 * @return the associated TR-069 engine or NULL if none available
 */
struct tr069_engine *frontend_get_tr069_engine (struct frontend *fe) ARGS_NONNULL;
#endif


/**
 * @brief Return the event base used by this frontend
 *
 * @param[in] fe    the frontend to get the used event base from
 *
 * @return the event base used by this frontend
 */
struct event_base *frontend_get_event_base (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Exit frontend after finishing all unfinished business
 *
 * @param[in] fe    the frontend which is supposed to exit
 */
void frontend_exit_when_idle (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Schedule the execution of an  out-of-session-event callback when no
 * 	  TR-069 session and/or USP communication is active
 *
 * @param[in] fe	the frontend the schedule the callback on
 * @param[in] cb	the callback to schedule
 * @param[in] data	the data to pass on to the callback upon execution
 */
void frontend_add_oose_callback (struct frontend *fe, oose_callback cb, void *data) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(2);


/**
 * @brief Process the out-of-session-event callbacks after finishing a TR-069 and/or USP
 * 	  session if all communication has concluded
 *
 * @param[in] fe    the frontend to process the scheduled OOSE callbacks on
 */
void frontend_process_oose (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Initialise some crucial datamodel parameters with useful values
 *
 * @param[in] fe	the frontend to initialise the datamodel parameters for
 * @param[in] oui	the OUI to use for this implementation
 * @param[in] pc	the product class to use for this implementation
 * @param[in] sn	the serial number to use for this implementation
 * @param[in] epid	the local USP endpoint identifier
 */
void frontend_init_datamodel_defaults (struct frontend *fe, char *oui, char *pc, char *sn, char *epid) ARGS_NONNULL;


/**
 * @brief Return the datamodel root object used by this frontend
 *
 * @param[in] fe	the frontend to get the used datamodel root object from
 *
 * @return the datamodel root object used by this frontend
 */
struct dm_object *frontend_get_dm_root (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Set "absolute time" flag, meaning the system time is somewhat recent to allow for e.g. certificate expiration checks.
 *
 * @param[in] fe	the frontend to set the absolute time flag on
 */
void frontend_set_absolutetime (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Get the "absolute time" flag.
 *
 * @param[in] fe	the frontend to get the absolute time flag from
 *
 * @return 1 if the system has obtained "absolute time", 0 if not
 */
int frontend_get_absolutetime (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Set the filename used for tracing of all unencryted HTTPish communication. DO NOT USE IN PRODUCTION!
 *
 * @param[in] fe	the frontend to set the HTTP log file on
 * @param[in] http_log	a filename path, can be NULL do disable the HTTP logging
 */
void frontend_set_http_logfilename (struct frontend *fe, char *http_log) ARGINDEX_NONNULL(1);


/**
 * @brief Retrieve the filename used for tracing of all HTTPish communication.
 *
 * @param[in] fe	the frontend to get the filename path for
 *
 * @return a filename path to the file to use or NULL if disabled/not configured
 */
char *frontend_get_http_logfilename (struct frontend *fe) ARGS_NONNULL;


#if defined(TR069_XMPP_SUPPORT)
/**
 * @brief Set the filename used for tracing of all unencryted XMPP communication. DO NOT USE IN PRODUCTION!
 *
 * @param[in] fe	the frontend to set the xmpp log file on
 * @param[in] xmpp_log	a filename path, can be NULL do disable the xmpp logging
 */
void frontend_set_xmpp_logfilename (struct frontend *fe, char *xmpp_log) ARGINDEX_NONNULL(1);


/**
 * @brief Retrieve the filename used for tracing of all XMPP communication.
 *
 * @param[in] fe	the frontend to get the filename path for
 *
 * @return a filename path to the file to use or NULL if disabled/not configured
 */
char *frontend_get_xmpp_logfilename (struct frontend *fe) ARGS_NONNULL;
#endif


/**
 * @brief Add a notification subscriber to the notification list of the frontend
 *
 * @param[in] fe	the frontend to add the subscriber to
 * @param[in] cb	the callback to execute for notification events
 * @param[in] data	additional data to pass to subscriber when callback is executed
 */
void frontend_add_subscriber (struct frontend *fe, notification_cb cb, void *data) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(2);


/**
 * @brief Add a new notification to the list managed by this frontend
 *
 * @param[in] fe	the frontend add the notification to
 * @param[in] type	the type of notification to add
 * @param[in] data	optional data for the notification
 */
void frontend_add_notification (struct frontend *fe, enum notification_type type, void *data) ARGINDEX_NONNULL(1);


/**
 * @brief Process notifications by distributing them to the known engines
 *
 * @param[in] fe	the frontend to process the notifications for
 */
void frontend_process_notifications (struct frontend *fe);


/**
 * @brief Run endless event processing loop until we're ready to exit
 *
 * @param[in] fe	the frontend to process events for endlessly
 */
void frontend_run_forever (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Run event processing loop once processing all ready events unless exit was requested
 *
 * @param[in] fe	the frontend to process ready events for once
 *
 * @return 1 if the frontend is supposed to continue processing or 0 if the frontend ready for shutdown
 */
int frontend_run_once (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Run event processing loop once processing all ready events
 *
 * @param[in] fe	the frontend to process ready events for once
 */
void frontend_force_run_once (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Prepare everything frontend related for a graceful shutdown
 *
 * @param[in] fe	the frontend to shutdown
 */
void frontend_shutdown (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Free the frontend and all its associated data
 *
 * @param[in] fe	the frontend to free
 */
void frontend_free (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Starts a new data model session
 *
 * @param[in] fe	the frontend on which to start a new data model session
 */
void frontend_session_start (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Resets a started data model session
 *
 * @param[in] fe	the frontend on which to reset new data model session
 */
void frontend_session_reset (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Ends a data model session
 *
 * @param[in] fe	the frontend on which to end a data model session
 */
void frontend_session_done (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Immediately updates the parameter value and notifies everyone about this
 *
 * @param[in] fe	pointer to frontend
 * @param[in] par	pointer to parameter to be updated
 * @param[in] value	the new parameter value to be set
 *
 * @return 0 if value change was successful or != 0 otherwise
 */
int frontend_set_parameter_value (struct frontend *fe, struct dm_parameter *par, char *value) ARGS_NONNULL;


#if defined(ENABLE_TR069)
/**
 * @brief Retrieve the Device.ManagementServer. object
 *
 * @param[in] fe	a pointer to the frontend
 *
 * @return a pointer to the Device.ManagementServer. object
 */
struct dm_object *frontend_get_managementserver_object (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;
#endif


/**
 * @brief Retrieve the Device.DeviceInfo. object
 *
 * @param[in] fe	a pointer to the frontend
 *
 * @return a pointer to the Device.DeviceInfo. object
 */
struct dm_object *frontend_get_deviceinfo_object (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


#if defined(ENABLE_USP)
/**
 * @brief Retrieve the Device.LocalAgent. object
 *
 * @param[in] fe	a pointer to the frontend
 *
 * @return a pointer to the Device.LocalAgent. object
 */
struct dm_object *frontend_get_localagent_object (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;
#endif


/**
 * @brief Execute the tabinit callback function for all the table roots from the root object down the path
 *        if they haven't been updated before in this session
 *
 * @param[in] path The path to traverse from root down. This MUST be a valid full path.
 *
 * @return A CWMP error code
 */
unsigned int frontend_update_tables_in_path (struct frontend *fe, char *path) ARGS_NONNULL;


#endif /* CORE_FRONTEND_H_ */
