/**
 * @file
 *
 * @copyright Axiros GmbH 2021, All rights reserved
 *
 * @brief Definition of datamodel permission handling functionality
 */

#ifndef DMPERMISSION_H
#define DMPERMISSION_H

#include "compiler.h"

/* A type for permissions sets (nxRWXN) on parameters and objects */
typedef unsigned int dm_permission_t;


/* A type for single set of permissions on parameters and objects in the form RWXN */
typedef unsigned int dm_single_permission_t;


/* The maximum index of permissions available for use in a permission set */
#define DM_PERMISSIONS_MAX_INDEX (sizeof (dm_permission_t) * 2 - 1)


/**
 * @brief Yield the default permission set which is "all allowed for anyone"
 *
 * @return The default permission set
 */
dm_permission_t dm_permission_default (void);


/**
 * @brief Yield the permission of the nth group in the given permission set
 *
 * @param[in] perms	The permission set to extract the single permission from
 * @param[in] nth	The group index (zero indexed) to extract the single permissions for
 *
 * @return The permissions of the nth group in the permission set or 0xF if nth > DM_PERMISSION_MAX_INDEX
 */
dm_single_permission_t dm_permission_get_nth (dm_permission_t perms, unsigned char nth);


/**
 * @brief Adjust the permissions of the nth group according to the parameters and return the new permission set
 *
 * @param[in] perms	The permission set to merge the given single permission into
 * @param[in] nth	The group index (zero indexed) to set the single permissions for
 * @param[in] single	The single group permissions to merge into the permission set
 *
 * @return The merged permissions set, will result in a no operation if nth > DM_PERMISSION_MAX_INDEX
 */
dm_permission_t dm_permission_set_nth (dm_permission_t perms, unsigned char nth, dm_single_permission_t single);


/**
 * @brief Return whether the permissions include the execute permission
 *
 * @param[in] single	The single group permissions to return the executable permission state on
 *
 * @return 1 if execute is permitted, 0 if execute is forbidden
 */
unsigned int dm_permission_get_single_executable (dm_single_permission_t single);


/**
 * @brief Adjust the execute permission flag of the single group permission
 *
 * @param[in] single	The single group permissions to set the execute permission flag in
 * @param[in] flag	The flag to set, 0 for execute forbidden, != 0 for execute permitted
 *
 * @return The modified single group permissions
 */
dm_single_permission_t dm_permission_set_single_executable (dm_single_permission_t single, unsigned int flag);


/**
 * @brief Return whether the permissions include the notify permission
 *
 * @param[in] single	The single group permissions to return the notifyable permission state on
 *
 * @return 1 if notify is permitted, 0 if notify is forbidden
 */
unsigned int dm_permission_get_single_notifyable (dm_single_permission_t single);


/**
 * @brief Adjust the notify permission flag of the single group permission
 *
 * @param[in] single	The single group permissions to set the notify permission flag in
 * @param[in] flag	The flag to set, 0 for notify forbidden, != 0 for notify permitted
 *
 * @return The modified single group permissions
 */
dm_single_permission_t dm_permission_set_single_notifyable (dm_single_permission_t single, unsigned int flag);


/**
 * @brief Return whether the permissions include the read permission
 *
 * @param[in] single	The single group permissions to return the readable permission state on
 *
 * @return 1 if read is permitted, 0 if read is forbidden
 */
unsigned int dm_permission_get_single_readable (dm_single_permission_t single);


/**
 * @brief Adjust the read permission flag of the single group permission
 *
 * @param[in] single	The single group permissions to set the read permission flag in
 * @param[in] flag	The flag to set, 0 for read forbidden, != 0 for read permitted
 *
 * @return The modified single group permissions
 */
dm_single_permission_t dm_permission_set_single_readable (dm_single_permission_t single, unsigned int flag);


/**
 * @brief Return whether the permissions include the write permission
 *
 * @param[in] single	The single group permissions to return the writeable permission state on
 *
 * @return 1 if write is permitted, 0 if write is forbidden
 */
unsigned int dm_permission_get_single_writeable (dm_single_permission_t single);


/**
 * @brief Adjust the write permission flag of the single group permission
 *
 * @param[in] single	The single group permissions to set the write permission flag in
 * @param[in] flag	The flag to set, 0 for write forbidden, != 0 for write permitted
 *
 * @return The modified single group permissions
 */
dm_single_permission_t dm_permission_set_single_writeable (dm_single_permission_t single, unsigned int flag);


/**
 * @brief Parse an exactly 4 byte long '\0' terminated string into a permission group
 *
 * @param[in] perms	The permissions string according to TR-181i2a14 (cf. Device.LocalAgent.ControllerTrust.Role.{i}.Permission.{i}.)
 *
 * @note The string MUST be exactly 4 bytes long, in the order "rwxn" (all lowercase), all other characters (also in incorrect order) will treated as a '-'
 *
 * @return The single group permissions matching the textual description
 */
dm_single_permission_t dm_permission_parse_single (char *perms) ARGS_NONNULL;


/**
 * @brief Create an allocated string explaining the permissions in the specified permission set
 *
 * @param[in] perms	The permission set to explain
 *
 * @return A freshly allocated string with 4 characters per group either containing r, w, n, x or - in the respective positions
 */
char *dm_permission_explain (dm_permission_t perms) RETURNS_NONNULL MALLOC;

#endif /* DMPERMISSION_H */
